@@ -2,10 +2,10 @@
 #include "config.h"
 #endif
 
-#include <sys/ttydefaults.h>
 #include <ctype.h>
 #include <err.h>
 #include <locale.h>
+#include <poll.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -28,14 +28,11 @@
 #define EX_SIG 128
 #define EX_SIGINT (EX_SIG + SIGINT)
 
-#define MAX(x, y) ((x) > (y) ? (x) : (y))
-#define tty_putp(capability) do { \
-	if (tputs(capability, 1, tty_putc) == ERR) \
-		errx(1, #capability ": unknown terminfo capability"); \
+#define tty_putp(capability) do {					\
+	if (tputs(capability, 1, tty_putc) == ERR)			\
+		errx(1, #capability ": unknown terminfo capability");	\
 	} while (0)
 
-#define ESCAPE 27
-
 enum {
 	UNKNOWN,
 	ALT_ENTER,
@@ -50,69 +47,62 @@ enum {
 	UP,
 	RIGHT,
 	DOWN,
-	LEFT
+	LEFT,
+	PAGE_DOWN,
+	PAGE_UP
 };
 
 struct choice {
 	char	*description;
 	char	*string;
 	size_t	 length;
-	size_t	 printable_length;
-	size_t	 match_start;
-	size_t	 match_end;
+	ssize_t	 match_start;	/* inclusive match start offset */
+	ssize_t	 match_end;	/* exclusive match end offset */
 	float	 score;
 };
 
-__dead static void		 usage(void);
-__dead static void		 version(void);
-static void			 get_choices(void);
-static char			*eager_strpbrk(const char *, const char *);
-static size_t			 count_printable(char *, size_t length);
-static void			 put_choice(const struct choice *);
-static const struct choice	*selected_choice(void);
-static void			 filter_choices(void);
-static void			 score(struct choice *);
-static int			 min_match(const char *, size_t, size_t *, size_t *);
-static char			*strcasechr(const char *, const char *);
-static void			 init_tty(void);
-static int			 tty_putc(int);
-static void			 handle_sigint(int);
-static void			 restore_tty(void);
-static void			 put_line(char *, int);
-static int			 print_choices(int);
-static int			 get_key(char *, size_t, size_t *);
-static int			 tty_getc(void);
-static void			 delete_between(char *, size_t, size_t, size_t);
-static void			 print_query(char *, size_t, size_t, size_t);
 static int			 choicecmp(const void *, const void *);
+static void			 delete_between(char *, size_t, size_t, size_t);
+static char			*eager_strpbrk(const char *, const char *);
+static void			 filter_choices(void);
+static void			 get_choices(void);
+static int			 get_key(char *, size_t, size_t *);
+static void			 handle_sigint(int);
 static int			 isu8cont(unsigned char);
 static int			 isu8start(unsigned char);
+static int			 min_match(const char *, size_t, ssize_t *,
+				    ssize_t *);
+static int			 print_choices(int, int);
+static void			 print_line(const char *, size_t, int, ssize_t,
+				    ssize_t);
+static void			 score(struct choice *);
+static const struct choice	*selected_choice(void);
+static const char		*strcasechr(const char *, const char *);
+static int			 tty_getc(void);
+static void			 tty_init(void);
+static int			 tty_putc(int);
+static void			 tty_restore(void);
+__dead static void		 usage(void);
 
-static FILE		*tty_in;
-static FILE		*tty_out;
-static char		*query;
-static int		 descriptions;
-static int		 output_description;
-static int		 use_alternate_screen = 1;
-static int		 sort = 1;
-static size_t		 query_size;
 static struct termios	 original_attributes;
 static struct {
 	size_t size;
 	size_t length;
 	struct choice *v;
-} choices;
-static struct {
-	size_t	size;
-	size_t	length;
-	char	*string;
-} input;
+}			 choices;
+static FILE		*tty_in, *tty_out;
+static char		*query;
+static size_t		 query_length, query_size;
+static int		 descriptions, choices_lines;
+static int		 sort = 1;
+static int		 use_alternate_screen = 1;
 
 int
 main(int argc, char **argv)
 {
 	const struct choice	*choice;
-	int			 option;
+	int			 c;
+	int			 output_description = 0;
 
 #ifdef HAVE_PLEDGE
 	if (pledge("stdio tty rpath wpath cpath", NULL) == -1)
@@ -121,8 +111,8 @@ main(int argc, char **argv)
 
 	setlocale(LC_CTYPE, "");
 
-	while ((option = getopt(argc, argv, "dhoq:SvxX")) != -1) {
-		switch (option) {
+	while ((c = getopt(argc, argv, "dhoq:SvxX")) != -1)
+		switch (c) {
 		case 'd':
 			descriptions = 1;
 			break;
@@ -136,13 +126,15 @@ main(int argc, char **argv)
 		case 'q':
 			if ((query = strdup(optarg)) == NULL)
 				err(1, "strdup");
-			query_size = strlen(query) + 1;
+			query_length = strlen(query);
+			query_size = query_length + 1;
 			break;
 		case 'S':
 			sort = 0;
 			break;
 		case 'v':
-			version();
+			puts(PACKAGE_VERSION);
+			exit(EX_OK);
 		case 'x':
 			use_alternate_screen = 1;
 			break;
@@ -152,7 +144,6 @@ main(int argc, char **argv)
 		default:
 			usage();
 		}
-	}
 	argc -= optind;
 	argv += optind;
 	if (argc > 0)
@@ -165,7 +156,7 @@ main(int argc, char **argv)
 	}
 
 	get_choices();
-	init_tty();
+	tty_init();
 
 #ifdef HAVE_PLEDGE
 	if (pledge("stdio tty", NULL) == -1)
@@ -173,12 +164,14 @@ main(int argc, char **argv)
 #endif
 
 	choice = selected_choice();
-	restore_tty();
-	if (choice != NULL)
-		put_choice(choice);
+	tty_restore();
+	if (choice != NULL) {
+		printf("%s\n", choice->string);
+		if (output_description)
+			printf("%s\n", choice->description);
+	}
 
 	free(choices.v);
-	free(input.string);
 	free(query);
 
 	return EX_OK;
@@ -200,45 +193,40 @@ usage(void)
 	exit(EX_USAGE);
 }
 
-__dead void
-version(void)
-{
-	puts(PACKAGE_VERSION);
-	exit(EX_OK);
-}
-
 void
 get_choices(void)
 {
-	char		*description, *field_separators, *start, *stop;
-	ssize_t		 length;
+	char		*buf, *description, *field_separators, *start, *stop;
+	ssize_t		 n;
+	size_t		 length = 0;
+	size_t		 size = BUFSIZ;
 
 	if ((field_separators = getenv("IFS")) == NULL)
 		field_separators = " ";
 
-	input.size = BUFSIZ;
-	if ((input.string = malloc(input.size)) == NULL)
+	if ((buf = malloc(size)) == NULL)
 		err(1, NULL);
 	for (;;) {
-		if ((length = read(STDIN_FILENO, input.string + input.length,
-				   input.size - input.length)) <= 0)
+		if ((n = read(STDIN_FILENO, buf + length, size - length)) == -1)
+			err(1, "read");
+		else if (n == 0)
 			break;
 
-		input.length += length;
-		if (input.length + 1 < input.size)
+		length += n;
+		if (length + 1 < size)
 			continue;
-		input.size *= 2;
-		if ((input.string = realloc(input.string, input.size)) == NULL)
+		if ((buf = reallocarray(buf, 2, size)) == NULL)
 			err(1, NULL);
+		size *= 2;
 	}
-	memset(input.string + input.length, '\0', input.size - input.length);
+	memset(buf + length, '\0', size - length);
 
 	choices.size = 16;
 	if ((choices.v = reallocarray(NULL, choices.size,
 			    sizeof(struct choice))) == NULL)
 		err(1, NULL);
 
-	start = input.string;
+	start = buf;
 	while ((stop = strchr(start, '\n')) != NULL) {
 		*stop = '\0';
 
@@ -251,10 +239,8 @@ get_choices(void)
 		choices.v[choices.length].length = stop - start;
 		choices.v[choices.length].string = start;
 		choices.v[choices.length].description = description;
-		choices.v[choices.length].printable_length =
-		    count_printable(start, stop - start);
-		choices.v[choices.length].match_start = 0;
-		choices.v[choices.length].match_end = 0;
+		choices.v[choices.length].match_start = -1;
+		choices.v[choices.length].match_end = -1;
 		choices.v[choices.length].score = 0;
 
 		start = stop + 1;
@@ -272,7 +258,8 @@ get_choices(void)
 char *
 eager_strpbrk(const char *string, const char *separators)
 {
-	char	*ptr = NULL, *tmp_ptr;
+	char	*tmp_ptr;
+	char	*ptr = NULL;
 
 	for (tmp_ptr = strpbrk(string, separators);
 	     tmp_ptr;
@@ -282,70 +269,68 @@ eager_strpbrk(const char *string, const char *separators)
 	return ptr;
 }
 
-size_t
-count_printable(char *str, size_t length)
-{
-	size_t	i;
-	int	in_esc_seq = 0, printable = 0;
-
-	for (i = 0; i < length; i++) {
-		if (in_esc_seq) {
-			in_esc_seq = (str[i - 1] == ESCAPE && str[i] == '[') ||
-			    str[i] < '@' || str[i] > '~';
-		} else {
-			in_esc_seq = str[i] == ESCAPE;
-			if (!in_esc_seq)
-				printable++;
-		}
-	}
-	return printable;
-}
-
-void
-put_choice(const struct choice *choice)
-{
-	puts(choice->string);
-
-	if (output_description)
-		puts(choice->description);
-}
-
 const struct choice *
 selected_choice(void)
 {
-	size_t		cursor_position, length, query_length, scroll;
-	char		buf[6];
-	int		key, selection = 0, visible_choices_count;
-	int		word_position;
+	size_t	cursor_position, i, j, length;
+	size_t	xscroll = 0;
+	char	buf[6];
+	int	choices_count, key, word_position;
+	int	selection = 0;
+	int	yscroll = 0;
 
-	cursor_position = query_length = strlen(query);
+	cursor_position = strlen(query);
 
 	filter_choices();
 
-	if (cursor_position >= (size_t)columns)
-		scroll = cursor_position - columns + 1;
-	else
-		scroll = 0;
-
-	visible_choices_count = print_choices(selection);
-	print_query(query, query_length, cursor_position, scroll);
-	tty_putp(cursor_normal);
-
 	for (;;) {
+		tty_putp(cursor_invisible);
+		tty_putp(carriage_return);	/* move cursor to first column */
+		if (cursor_position >= xscroll + columns)
+			xscroll = cursor_position - columns + 1;
+		if (cursor_position < xscroll)
+			xscroll = cursor_position;
+		print_line(&query[xscroll], query_length - xscroll, 0, -1, -1);
+		choices_count = print_choices(yscroll, selection);
+		if ((size_t)choices_count - yscroll < choices.length
+		    && choices_count - yscroll < choices_lines) {
+			/*
+			 * Printing the choices did not consume all available
+			 * lines and there could still be choices left from the
+			 * last print in the lines not yet consumed.
+			 *
+			 * The clr_eos capability clears the screen from the
+			 * current column to the end. If the last visible choice
+			 * is selected, the standout in the last and current
+			 * column will be also be cleared. Therefore, move down
+			 * one line before clearing the screen.
+			 */
+			if (tty_putc('\n') == EOF)
+				err(1, "tty_putc");
+			tty_putp(clr_eos);
+			tty_putp(tparm(parm_up_cursor, choices_count + 1));
+		} else {
+			tty_putp(tparm(parm_up_cursor,
+				    choices_count < choices_lines
+				    ? choices_count : choices_lines));
+		}
+		tty_putp(carriage_return);	/* move cursor to first column */
+		for (i = j = 0; i < cursor_position; j++)
+			while (isu8cont(query[++i]));
+		if (j > 0)
+			/*
+			 * parm_right_cursor interprets 0 as 1, therefore only
+			 * move the cursor if the position is non zero.
+			 */
+			tty_putp(tparm(parm_right_cursor, j));
+		tty_putp(cursor_normal);
 		fflush(tty_out);
-		memset(buf, 0, sizeof(buf));
-		key = get_key(buf, sizeof(buf), &length);
 
+		key = get_key(buf, sizeof(buf), &length);
 		switch (key) {
 		case ENTER:
-			if (visible_choices_count > 0) {
-				if (selection >= 0
-				    && selection < (ssize_t)choices.length)
-					return &choices.v[selection];
-				else
-					return NULL;
-			}
-
+			if (choices_count > 0)
+				return &choices.v[selection];
 			break;
 		case ALT_ENTER:
 			choices.v[choices.length].string = query;
@@ -364,9 +349,8 @@ selected_choice(void)
 				cursor_position -= length;
 				query_length -= length;
 				filter_choices();
-				selection = 0;
+				selection = yscroll = 0;
 			}
-
 			break;
 		case DEL:
 			if (cursor_position < query_length) {
@@ -380,9 +364,8 @@ selected_choice(void)
 				    cursor_position + length);
 				query_length -= length;
 				filter_choices();
-				selection = 0;
+				selection = yscroll = 0;
 			}
-
 			break;
 		case CTRL_U:
 			delete_between(
@@ -393,7 +376,7 @@ selected_choice(void)
 			query_length -= cursor_position;
 			cursor_position = 0;
 			filter_choices();
-			selection = 0;
+			selection = yscroll = 0;
 			break;
 		case CTRL_K:
 			delete_between(
@@ -403,7 +386,7 @@ selected_choice(void)
 			    query_length);
 			query_length = cursor_position;
 			filter_choices();
-			selection = 0;
+			selection = yscroll = 0;
 			break;
 		case CTRL_W:
 			if (cursor_position == 0)
@@ -425,7 +408,7 @@ selected_choice(void)
 			query_length -= cursor_position - word_position;
 			cursor_position = word_position;
 			filter_choices();
-			selection = 0;
+			selection = yscroll = 0;
 			break;
 		case CTRL_A:
 			cursor_position = 0;
@@ -434,12 +417,18 @@ selected_choice(void)
 			cursor_position = query_length;
 			break;
 		case DOWN:
-			if (selection < visible_choices_count - 1)
-				++selection;
+			if (selection < choices_count - 1) {
+				selection++;
+				if (selection - yscroll == choices_lines)
+					yscroll++;
+			}
 			break;
 		case UP:
-			if (selection > 0)
-				--selection;
+			if (selection > 0) {
+				selection--;
+				if (selection - yscroll < 0)
+					yscroll--;
+			}
 			break;
 		case LEFT:
 			while (cursor_position > 0
@@ -449,6 +438,18 @@ selected_choice(void)
 			while (cursor_position < query_length
 			    && isu8cont(query[++cursor_position]));
 			break;
+		case PAGE_DOWN:
+			if (selection + choices_lines < choices_count)
+				yscroll = selection += choices_lines;
+			else
+				selection = choices_count - 1;
+			break;
+		case PAGE_UP:
+			if (selection - choices_lines > 0)
+				yscroll = selection -= choices_lines;
+			else
+				yscroll = selection = 0;
+			break;
 		default:
 			if (!isu8start(buf[0]) && !isprint(buf[0]))
 				continue;
@@ -470,29 +471,41 @@ selected_choice(void)
 			query_length += length;
 			query[query_length] = '\0';
 			filter_choices();
-			selection = 0;
+			selection = yscroll = 0;
 		}
-
-		tty_putp(cursor_invisible);
-
-		visible_choices_count = print_choices(selection);
-		if (cursor_position >= scroll + columns)
-			scroll = cursor_position - columns + 1;
-		if (cursor_position < scroll)
-			scroll = cursor_position;
-		print_query(query, query_length, cursor_position, scroll);
-		tty_putp(cursor_normal);
 	}
 }
 
+/*
+ * Filter choices using the current query while there is no new user input
+ * available.
+ */
 void
 filter_choices(void)
 {
-	int	i;
+	struct pollfd	pfd;
+	size_t		i;
+	int		nready;
 
-	for (i = 0; i < (ssize_t)choices.length; ++i)
+	for (i = 0; i < choices.length; i++) {
 		score(&choices.v[i]);
 
+		/*
+		 * Regularly check if there is any new user input available. If
+		 * true, abort filtering since the currently used query is
+		 * outdated. This improves the performance when the cardinality
+		 * of the choices is large.
+		 */
+		if (i > 0 && i % 50 == 0) {
+			pfd.fd = fileno(tty_in);
+			pfd.events = POLLIN;
+			if ((nready = poll(&pfd, 1, 0)) == -1)
+				err(1, "poll");
+			if (nready == 1 && pfd.revents & (POLLIN | POLLHUP))
+				break;
+		}
+	}
+
 	qsort(choices.v, choices.length, sizeof(struct choice), choicecmp);
 }
 
@@ -512,11 +525,12 @@ choicecmp(const void *p1, const void *p2)
 void
 score(struct choice *choice)
 {
-	size_t	query_length, match_length;
+	size_t	match_length;
 
 	if (min_match(choice->string, 0,
 		      &choice->match_start, &choice->match_end) == 0) {
-		choice->match_start = choice->match_end = choice->score = 0;
+		choice->match_start = choice->match_end = -1;
+		choice->score = 0;
 		return;
 	}
 
@@ -526,14 +540,14 @@ score(struct choice *choice)
 	}
 
 	match_length = choice->match_end - choice->match_start;
-	query_length = strlen(query);
 	choice->score = (float)query_length / match_length / choice->length;
 }
 
 int
-min_match(const char *string, size_t offset, size_t *start, size_t *end)
+min_match(const char *string, size_t offset, ssize_t *start, ssize_t *end)
 {
-	char	*s, *e, *q;
+	const char	*e, *q, *s;
+	size_t		 length;
 
 	q = query;
 	if ((s = e = strcasechr(&string[offset], q)) == NULL)
@@ -547,11 +561,11 @@ min_match(const char *string, size_t offset, size_t *start, size_t *end)
 			return 0;
 	}
 
+	length = e - s;
 	*start = s - string;
 	*end = e - string;
-	/* Less than or equal is used in order to obtain the left-most match. */
-	if (min_match(string, offset + 1, start, end)
-	    && (size_t)(e - s) <= *end - *start) {
+	if (length > query_length && min_match(string, offset + 1, start, end)
+	    && length < (size_t)(*end - *start)) {
 		*start = s - string;
 		*end = e - string;
 	}
@@ -562,7 +576,7 @@ min_match(const char *string, size_t offset, size_t *start, size_t *end)
  * Returns a pointer to first occurrence of the first character in s2 in s1 with
  * respect to Unicode characters and disregarding case.
  */
-char *
+const char *
 strcasechr(const char *s1, const char *s2)
 {
 	wchar_t	 wc1, wc2;
@@ -581,17 +595,16 @@ strcasechr(const char *s1, const char *s2)
 			continue;
 		}
 		if (wcsncasecmp(&wc1, &wc2, 1) == 0)
-			return (char *)s1;
+			return s1;
 	}
 
 	return NULL;
 }
 
 void
-init_tty(void)
+tty_init(void)
 {
 	struct termios	new_attributes;
-	int		i;
 
 	if ((tty_in = fopen("/dev/tty", "r")) == NULL)
 		err(1, "fopen");
@@ -606,16 +619,11 @@ init_tty(void)
 
 	setupterm((char *)0, fileno(tty_out), (int *)0);
 
+	choices_lines = lines - 1; /* available lines, minus query line */
+
 	if (use_alternate_screen)
 		tty_putp(enter_ca_mode);
 
-	/* Emit enough lines to fit all choices. */
-	for (i = 0; i < (ssize_t)choices.length && i < lines - 1; ++i)
-		tty_putp(cursor_down);
-	for (; i > 0; --i)
-		tty_putp(cursor_up);
-	tty_putp(save_cursor);
-
 	signal(SIGINT, handle_sigint);
 }
 
@@ -628,17 +636,17 @@ tty_putc(int c)
 void
 handle_sigint(int sig __attribute__((unused)))
 {
-	restore_tty();
+	tty_restore();
 	exit(EX_SIGINT);
 }
 
 void
-restore_tty(void)
+tty_restore(void)
 {
 	tcsetattr(fileno(tty_in), TCSANOW, &original_attributes);
 	fclose(tty_in);
 
-	tty_putp(restore_cursor);
+	tty_putp(carriage_return);	/* move cursor to first column */
 	tty_putp(clr_eos);
 
 	if (use_alternate_screen)
@@ -648,92 +656,86 @@ restore_tty(void)
 }
 
 void
-put_line(char *string, int length)
+print_line(const char *string, size_t length, int so, ssize_t ulso, ssize_t uleo)
 {
-	int	i;
+	size_t	i;
+	int	c, col, tabwidth;
+	int	in_esc_seq = 0;
+	int	non_printable = 0;
 
-	for (i = 0; i < length && i < columns; ++i) {
-		if (tty_putc(string[i]) == EOF)
-			err(1, "tty_putc");
-	}
-	for (; i < columns; ++i) {
-		if (tty_putc(' ') == EOF)
-			err(1, "tty_putc");
-	}
-}
+	if (so)
+		tty_putp(enter_standout_mode);
 
-void
-print_query(char *string, size_t length, size_t position, size_t scroll)
-{
-	size_t	i = 0;
+	for (col = i = 0; i < length && col < columns; i++) {
+		if (i == (size_t)ulso)
+			tty_putp(enter_underline_mode);
 
-	tty_putp(restore_cursor);
-	put_line(string + scroll, length - scroll);
-
-	tty_putp(restore_cursor);
-	while (i < position - scroll) {
-		while (isu8cont(string[++i]));
-		tty_putp(cursor_right);
-	}
-}
-
-int
-print_choices(int selection)
-{
-	struct choice	*choice;
-	size_t		 query_length;
-	int		 col, i, j, k;
-
-	tty_putp(clr_eos);
-	/* Emit query line. */
-	tty_putc('\n');
-
-	query_length = strlen(query);
-	for (choice = choices.v, col = i = 0;
-	     i < (ssize_t)choices.length
-	     && i < lines - 1
-	     && (query_length == 0 || choice->score > 0);
-	     choice++, i++) {
-		if (i == selection)
-			tty_putp(enter_standout_mode);
-
-		for (col = j = 0;
-		     j < (ssize_t)choice->match_start && col < columns;
-		     col += !isu8cont(choice->string[j]), j++)
-			if (tty_putc(choice->string[j]) == EOF)
-				err(1, "tty_putc");
-
-		tty_putp(enter_underline_mode);
-
-		for (;
-		     j < (ssize_t)choice->match_end && col < columns;
-		     col += !isu8cont(choice->string[j]), j++)
-			if (tty_putc(choice->string[j]) == EOF)
-				err(1, "tty_putc");
-
-		tty_putp(exit_underline_mode);
-
-		for (;
-		     j < (ssize_t)choice->length && col < columns;
-		     col += !isu8cont(choice->string[j]), j++) {
-			/* A null character will be present before the
-			 * terminating null character if descriptions is
-			 * enabled. */
-			if (choice->string[j] == '\0') {
-				if (tty_putc(' ') == EOF)
-					err(1, "tty_putc");
-			} else if (tty_putc(choice->string[j]) == EOF) {
-				err(1, "tty_putc");
-			}
+		/*
+		 * Count the number of outputted ANSI escape sequences
+		 * in order to adjust the column count since they do not
+		 * occupy any screen real-estate.
+		 */
+		if (in_esc_seq) {
+			non_printable++;
+			if (string[i] >= '@' && string[i] <= '~')
+				in_esc_seq = 0;
+		} else if (i > 0 && string[i - 1] == '\033'
+		    && string[i] == '[') {
+			in_esc_seq = 1;
+			non_printable = 2;
 		}
 
-		for (k = MAX(columns - choice->printable_length +
-			    (choice->length - col), 0); k > 0; k--)
-			if (tty_putc(' ') == EOF)
+		c = string[i];
+		if (c == '\t') {
+			/* Ceil column count to multiple of 8. */
+			col += tabwidth = 8 - (col & 7);
+			while (tabwidth-- > 0)
+				if (tty_putc(' ') == ERR)
+					err(1, "tty_putc");
+		} else {
+			/*
+			 * A null character will be present prior the
+			 * terminating null character if descriptions is
+			 * enabled.
+			 */
+			if (c == '\0')
+				c = ' ';
+			if (!isu8cont(c))
+				col++;
+			if (tty_putc(c) == EOF)
 				err(1, "tty_putc");
+		}
 
-		if (i == selection)
-			tty_putp(exit_standout_mode);
+		if (i + 1 == (size_t)uleo)
+			tty_putp(exit_underline_mode);
+	}
+	for (col -= non_printable; col < columns; col++)
+		if (tty_putc(' ') == EOF)
+			err(1, "tty_putc");
+
+	if (so)
+		tty_putp(exit_standout_mode);
+}
+
+/*
+ * Output as many choices as possible starting from offset and return the number
+ * of choices with a positive score. If the query is empty, all choices are
+ * considered having a positive score.
+ */
+int
+print_choices(int offset, int selection)
+{
+	struct choice	*choice;
+	int		 i;
+
+	for (i = offset, choice = &choices.v[i];
+	     (size_t)i < choices.length
+	     && (query_length == 0 || choice->score > 0);
+	     choice++, i++) {
+		if (i - offset < lines - 1)
+			print_line(choice->string, choice->length,
+			    i == selection, choice->match_start,
+			    choice->match_end);
 	}
 
 	return i;
@@ -743,48 +745,46 @@ int
 get_key(char *buf, size_t size, size_t *nread)
 {
 	static struct {
-		union {
-			const char *s;
-			char c;
-		} input;
-		size_t length;
-		int key;
-	} keys[] = {
-		{ { (char *)10 },		1,	ENTER },
-		{ { (char *)127 },		1,	BACKSPACE },
-		{ { (char *)CTRL('A') },	1,	CTRL_A },
-		{ { (char *)CTRL('B') },	1,	LEFT },
-		{ { (char *)CTRL('D') },	1,	DEL },
-		{ { (char *)CTRL('E') },	1,	CTRL_E },
-		{ { (char *)CTRL('F') },	1,	RIGHT },
-		{ { (char *)CTRL('K') },	1,	CTRL_K },
-		{ { (char *)CTRL('N') },	1,	DOWN },
-		{ { (char *)CTRL('P') },	1,	UP },
-		{ { (char *)CTRL('U') },	1,	CTRL_U },
-		{ { (char *)CTRL('W') },	1,	CTRL_W },
-		{ { "\033\n" },			2,	ALT_ENTER },
-		{ { "\033[A" },			3,	UP },
-		{ { "\033OA" },			3,	UP },
-		{ { "\033[B" },			3,	DOWN },
-		{ { "\033OB" },			3,	DOWN },
-		{ { "\033[C" },			3,	RIGHT },
-		{ { "\033OC" },			3,	RIGHT },
-		{ { "\033[D" },			3,	LEFT },
-		{ { "\033OD" },			3,	LEFT },
-		{ { "\033[3~" },		4,	DEL },
-		{ { "\033O3~" },		4,	DEL },
-		{ { NULL },			0,	0 },
+		const char	*s;
+		size_t		 length;
+		int		 key;
+	}	keys[] = {
+		{ "\n",		1,	ENTER },
+		{ "\177",	1,	BACKSPACE },
+		{ "\001",	1,	CTRL_A },
+		{ "\002",	1,	LEFT },
+		{ "\004",	1,	DEL },
+		{ "\005",	1,	CTRL_E },
+		{ "\006",	1,	RIGHT },
+		{ "\013",	1,	CTRL_K },
+		{ "\016",	1,	DOWN },
+		{ "\020",	1,	UP },
+		{ "\025",	1,	CTRL_U },
+		{ "\027",	1,	CTRL_W },
+		{ "\033\n",	2,	ALT_ENTER },
+		{ "\033[A",	3,	UP },
+		{ "\033OA",	3,	UP },
+		{ "\033[B",	3,	DOWN },
+		{ "\033OB",	3,	DOWN },
+		{ "\033[C",	3,	RIGHT },
+		{ "\033OC",	3,	RIGHT },
+		{ "\033[D",	3,	LEFT },
+		{ "\033OD",	3,	LEFT },
+		{ "\033[3~",	4,	DEL },
+		{ "\033O3~",	4,	DEL },
+		{ "\033[6~",	4,	PAGE_DOWN },
+		{ "\033[5~",	4,	PAGE_UP },
+		{ NULL,		0,	0 },
 	};
-	const char	*key;
-	int		 i;
+	int	c, i;
 
 	*nread = 0;
 getc:
 	buf[(*nread)++] = tty_getc();
 	size--;
-	for (i = 0; keys[i].input.s; i++) {
-		key = keys[i].length > 1 ? keys[i].input.s : &keys[i].input.c;
-		if (*nread > keys[i].length || strncmp(buf, key, *nread))
+	for (i = 0; keys[i].s != NULL; i++) {
+		if (*nread > keys[i].length
+		    || strncmp(buf, keys[i].s, *nread) != 0)
 			continue;
 
 		if (*nread == keys[i].length)
@@ -795,14 +795,19 @@ getc:
 			goto getc;
 	}
 
-	if (*nread > 1 && buf[0] == '\033' && (buf[1] == '[' || buf[1] == 'O'))
+	if (*nread > 1 && buf[0] == '\033' && (buf[1] == '[' || buf[1] == 'O')) {
 		/*
 		 * A escape sequence which is not a supported key is being read.
-		 * Ensure the whole sequence is read.
+		 * Discard the rest of the sequence.
 		 */
-		while ((buf[*nread - 1] < '@' || buf[*nread - 1] > '~')
-		    && size-- > 0)
-			buf[(*nread)++] = tty_getc();
+		for (;;) {
+			c = tty_getc();
+			if (c >= '@' && c <= '~')
+				break;
+		}
+
+		return UNKNOWN;
+	}
 
 	if (!isu8start(buf[0]))
 		return UNKNOWN;
